"""Monitor task updates and save them to a database using Django's ORM"""

import os
import logging
import threading
import traceback
from queue import Queue, Empty
from signal import signal, SIGTERM

from celery import Celery

logger = logging.getLogger('monitor_task_events')


class TaskEventMonitor:

    def __init__(self, app: Celery):
        self.app = app
        self.state = app.events.State()
        self.receiver = None
        self.task_updates = Queue()
        self.save_tasks = True
        self.worker_thread = threading.Thread(
            target=self.save_task_updates,
            name='TaskEventMonitor worker'
        )

        from celery_pantry.settings import pantry_settings

        handler_class = pantry_settings.CUSTOM_HANDLER
        if handler_class:
            from celery_pantry.custom_handler import CustomPantryHandler
            assert issubclass(handler_class, CustomPantryHandler)
            self._custom_handler = handler_class(self.state)
        else:
            self._custom_handler = None

        signal(SIGTERM, self.stop)

    def capture(self):
        logger.info('Capturing task updates')
        self.worker_thread.start()

        try:
            with self.app.connection() as connection:
                self.receiver = self.app.events.Receiver(connection, handlers={
                    '*': self.process_event,
                })
                self.receiver.capture()
        except KeyboardInterrupt:
            pass
        finally:
            self.stop()

    def stop(self, *args):
        logger.info('Stopping worker...')
        self.save_tasks = False
        if self.receiver:
            self.receiver.should_stop = True
        if self._custom_handler:
            self._custom_handler.stop()

        self.worker_thread.join()
        logger.info('Worker stopped')

    def process_event(self, event):
        self.state.event(event)

        if event['type'].startswith('task-'):
            task = self.state.tasks.get(event['uuid'])
            self.task_updates.put(task.as_dict())
        else:
            task = None

        if self._custom_handler:
            # noinspection PyBroadException
            try:
                self._custom_handler.process_event(event, task)
            except Exception:  # noqa
                logger.warning('Exception in custom handler: %s',
                               traceback.format_exc())

    def save_task_updates(self):
        from celery_pantry.models import Task

        while self.save_tasks:
            # noinspection PyBroadException
            try:
                task = self.task_updates.get(timeout=2)
                logger.debug('task: %s', task)
                task_id = task.pop('uuid')
                if task['worker'] is not None:
                    task['worker'] = task['worker'].hostname

                try:
                    obj = Task.objects.get(id=task_id)
                    obj.data = task
                    obj.save()
                except Task.DoesNotExist:
                    Task.objects.create(id=task_id, data=task)

            except Empty:
                continue

            except Exception:  # noqa
                logger.warning('Exception in task update thread: %s',
                               traceback.format_exc())


def monitor(app):
    TaskEventMonitor(app).capture()


if __name__ == '__main__':
    import django

    django.setup()
    logging.basicConfig(
        level=getattr(logging, os.getenv('LOGGING_LEVEL', 'INFO')),
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )

    app = Celery(os.getenv('CELERY_APP_NAME', 'example_tasks'),
                 broker='amqp://{}:{}@{}//'.format(
                     os.getenv('MQ_USER', 'example'),
                     os.getenv('MQ_PASS', 'publicsecret'),
                     os.getenv('MQ_HOST', 'mq'),
                 ))
    config = os.getenv('CELERY_CONFIG_OBJECT')
    if config:
        app.config_from_object(config)
    monitor(app)



from django.contrib import admin
from django.contrib.admin import ModelAdmin

from celery_pantry import models


class TaskAdmin(ModelAdmin):

    """Read-only access to Task objects from admin panel"""

    readonly_fields = ('id', 'data')
    actions = None

    def has_delete_permission(self, request, obj=None):
        return False


admin.site.register(models.Task, TaskAdmin)



# Generated by Django 2.1.1 on 2018-09-16 14:20

import django.contrib.postgres.fields.jsonb
import django.contrib.postgres.indexes
from django.contrib.postgres.operations import BtreeGinExtension
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        BtreeGinExtension(),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.CharField(help_text='Task ID', max_length=512, primary_key=True, serialize=False)),
                ('first_saved', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(blank=True, help_text='Task data')),
            ],
        ),
        migrations.AddIndex(
            model_name='task',
            index=django.contrib.postgres.indexes.GinIndex(fields=['data'], name='celery_pant_data_785c1b_gin'),
        ),
    ]



import abc


class CustomPantryHandler(metaclass=abc.ABCMeta):

    def __init__(self, state):
        self.state = state

    @abc.abstractmethod
    def process_event(self, event, task):
        pass

    def stop(self):
        pass



from django.conf import settings
from django.core.signals import setting_changed
from rest_framework.settings import APISettings


class Settings(APISettings):

    SETTINGS_KEY = 'CELERY_PANTRY'

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, self.SETTINGS_KEY, {})
        return self._user_settings


DEFAULTS = {
    'CUSTOM_HANDLER': None
}
IMPORT_STRINGS = (
    'CUSTOM_HANDLER',
)
pantry_settings = Settings(None, DEFAULTS, IMPORT_STRINGS)


def reload_settings(*args, **kwargs):
    setting = kwargs['setting']
    if setting == Settings.SETTINGS_KEY:
        pantry_settings.reload()


setting_changed.connect(reload_settings)



from django.db import models
from django.contrib.postgres import fields, indexes


class Task(models.Model):

    """Celery task information"""

    class Meta:
        indexes = [
            indexes.GinIndex(fields=['data']),
        ]

    id = models.CharField(max_length=512, primary_key=True,
                          help_text='Task ID')
    first_saved = models.DateTimeField(auto_now_add=True, db_index=True)
    data = fields.JSONField(blank=True, help_text='Task data')

    def __str__(self):
        return 'Task {}'.format(self.id)



import os

import setuptools

import celery_pantry


classifiers = """
    Development Status :: 3 - Alpha
    License :: OSI Approved :: MIT License
    Topic :: System :: Archiving
    Topic :: System :: Monitoring
    Topic :: System :: Distributed Computing
    Topic :: Software Development :: Object Brokering
    Intended Audience :: Developers
    Intended Audience :: System Administrators
    Programming Language :: Python :: 3.6
    Programming Language :: Python :: Implementation :: CPython
    Operating System :: OS Independent
"""

with open('README.rst') as f:
    README = f.read()


def find_package_data(package_path, *data_paths):
    """Recursively list files in a directory for use in package_data."""
    start_dir = os.path.abspath('.')
    os.chdir(package_path)
    data = []
    try:
        for data_path in data_paths:
            for root, dirs, files in os.walk(data_path):
                for file_name in files:
                    data.append(os.path.join(root, file_name))
    finally:
        os.chdir(start_dir)
    return data


setuptools.setup(
    name='celery_pantry',
    packages=setuptools.find_packages(exclude=('example_project',)),
    version=celery_pantry.__version__,
    description='Monitor and archive Celery task information',
    long_description=README,
    keywords='celery monitoring archiving',
    author='Toni Ruža',
    author_email='toni.ruza@gmail.com',
    url='https://github.com/bosonogi/celery_pantry',
    license='MIT',
    platforms=['any'],
    install_requires=[
        "celery>=4.0",
        "Django>=2.0",
        "coreapi>=2.3",
        "djangorestframework>=3.8",
        "psycopg2-binary~=2.7",
    ],
    package_data={
        'celery_pantry': find_package_data('celery_pantry', 'templates')
    },
    python_requires=">=3.6",
    classifiers=[s.strip() for s in classifiers.splitlines() if s],
    zip_safe=True,
)



import os
import json

import coreapi
import coreschema
from django.template import loader
from rest_framework import viewsets, filters
from rest_framework.filters import OrderingFilter

from celery_pantry import models, serializers


class TasksFilter(filters.BaseFilterBackend):

    """Return tasks that satisfy given constraints"""

    template = os.path.join('celery_pantry', 'tasks_filter.html')
    schema = [
        coreapi.Field(
            name='with_data',
            required=False,
            location='query',
            schema=coreschema.Object(
                description='Data users must have',
                min_properties=1
            )
        ),
        coreapi.Field(
            name='without_data',
            required=False,
            location='query',
            schema=coreschema.Object(
                description='Data users must not have',
                min_properties=1
            )
        ),
    ]

    @staticmethod
    def _filter__with_data(queryset, value):
        return queryset.filter(data__contains=json.loads(value))

    @staticmethod
    def _filter__without_data(queryset, value):
        return queryset.exclude(data__contains=json.loads(value))

    def filter_queryset(self, request, queryset, view):
        if view.action == 'list':
            for key, value in request.query_params.items():
                if value:
                    try:
                        filter_func = getattr(self, f'_filter__{key}')
                        queryset = filter_func(queryset, value)
                    except AttributeError:
                        continue
        return queryset

    def to_html(self, request, queryset, view) -> str:
        template = loader.get_template(self.template)
        context = self.get_template_context(request)
        return template.render(context)

    def get_template_context(self, request):
        return dict(conditions={
            field.name: request.query_params.get(field.name, '')
            for field in self.schema
        })


class TaskViewSet(viewsets.ReadOnlyModelViewSet):

    """Celery Tasks"""

    queryset = models.Task.objects.all()
    serializer_class = serializers.TaskSerializer
    lookup_field = 'id'
    lookup_value_regex = '[a-z0-9._-]+'
    filter_backends = (OrderingFilter, TasksFilter)
    ordering_fields = ('first_saved',)
    ordering = ('-first_saved',)



from rest_framework import serializers
from rest_framework.settings import api_settings

from celery_pantry import models


class TaskSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = models.Task
        fields = (
            api_settings.URL_FIELD_NAME,
            'id',
            'first_saved',
            'data',
        )
        extra_kwargs = {
            api_settings.URL_FIELD_NAME: {'lookup_field': 'id'}
        }
